//! Storage layer trait
use crate::{
    amount::AmountCents,
    asset::AssetId,
    changelog::{self, Changelog},
    transaction::{from_db, Type},
    AccountId, Amount, Payment, PaymentId, Status, Transaction, TransactionId,
};
use serde::{de::DeserializeOwned, Serialize};

#[derive(thiserror::Error, Debug, Serialize)]
/// Storage error
pub enum Error {
    #[error("Storage error: {0}")]
    /// Any error related to the storage layer, with string
    Storage(String),

    #[error("Invalid date format: {0}")]
    /// The date format is invalid
    InvalidDate(String),

    #[error("Spend payment: {0}")]
    /// The requested payment has been spent already
    SpendPayment(String),

    #[error("No storage update when expecting")]
    /// No update was performed when expecting one
    NoUpdate,

    #[error("Record not found")]
    /// The requested record was not found
    NotFound,

    #[error("Error while parsing changelog: {0}")]
    /// Error while parsing changelog
    Changelog(#[from] changelog::Error),

    #[error("Not enough unspent payments (missing {0} cents)")]
    /// TODO: Convert the AmountCents to Amount for better error reporting upstream
    NotEnoughUnspentPayments(AmountCents),
}

#[async_trait::async_trait]
/// Batches
///
/// Batches are a group of updates to the databases, in which all of the are
/// executed or none. Same concept as a rdbms transaction.
pub trait Batch<'a> {
    /// The current batch is discarded
    async fn rollback(self) -> Result<(), Error>;

    /// The batch is commited into the storage layer and it is consumed.
    async fn commit(self) -> Result<(), Error>;

    /// Stores a changelogs event
    async fn store_changelogs<T: DeserializeOwned + Serialize + Send + Sync>(
        &mut self,
        changelog: &[Changelog<T>],
    ) -> Result<(), Error>;

    /// Spends a payment.
    ///
    /// Under normal circumstances the transaction_id is required, and the
    /// payment is marked as spent.
    ///
    /// If the transaction_id is None, the payment is released and can be spent
    /// again. This will only happen when the transaction that spent this
    /// payment was either cancelled or failed. The original transaction will
    /// exists forever either as failed or cancelled and will reference the
    /// original payment, but the underlying payment that was just released will
    /// become spendable again
    async fn update_payment(
        &mut self,
        payment_id: &PaymentId,
        spent_by: &TransactionId,
        spent_status: Status,
    ) -> Result<(), Error>;

    /// Returns the stats of a payment from the point of view of the on-going transaction
    async fn get_payment_status(
        &mut self,
        transaction_id: &TransactionId,
    ) -> Result<Option<Status>, Error>;

    /// Stores a payment (that was created by a transaction)
    async fn store_new_payment(&mut self, payment: &Payment) -> Result<(), Error>;

    /// Stores a transaction
    async fn store_transaction(&mut self, transaction: &Transaction) -> Result<(), Error>;

    /// Sets the tags for a given transaction. Any tag not included in this
    /// vector should be removed
    async fn tag_transaction(
        &mut self,
        transaction: &Transaction,
        tags: &[String],
    ) -> Result<(), Error>;

    /// Creates a relationship between an account and a transaction.
    ///
    /// This is a chance to build an index which relates accounts to
    /// transactions (spending or receiving accounts). This information is used
    /// to query ledger information per account and to enhance any future cache
    /// layer.
    async fn relate_account_to_transaction(
        &mut self,
        transaction: &Transaction,
        account: &AccountId,
    ) -> Result<(), Error>;
}

#[async_trait::async_trait]
/// Main storage layer
pub trait Storage<'a> {
    /// The batch trait.
    ///
    /// This batch is needed to perform the changes in the transactions and
    /// payments in a atomic way
    type Batch: Batch<'a>;

    /// Begins a transaction
    ///
    /// A transaction prepares the storage for a batch of operations, where all
    /// the operations are persisted or none.
    ///
    /// The transaction is returned as a Batch object, which is used to perform
    /// the changes in the transactions and payments.
    ///
    /// The batch has also a rollback
    async fn begin(&'a self) -> Result<Self::Batch, Error>;

    /// Returns a payment object by ID.
    async fn get_payment(&self, id: &PaymentId) -> Result<Payment, Error>;

    /// Similar to get_payment() but errors if the requested payment is not spendable.
    async fn get_unspent_payment(&self, id: &PaymentId) -> Result<Payment, Error> {
        let payment = self.get_payment(id).await?;
        if payment.is_spendable() {
            Ok(payment)
        } else {
            Err(Error::NotFound)
        }
    }

    /// Returns the balances for a given account
    async fn get_balance(&self, account: &AccountId) -> Result<Vec<Amount>, Error>;

    /// Returns a list of unspent payments for a given account and asset.
    ///
    /// The payments should be returned sorted by ascending amount, this bit is
    /// important to make sure that all negative payments are always taken into
    /// account. It will also improve the database by using many small payments
    /// instead of a few large ones, which will make the database faster leaving
    /// fewer unspent payments when checking for balances.
    async fn get_unspent_payments(
        &self,
        account: &AccountId,
        asset: AssetId,
        target_amount: AmountCents,
    ) -> Result<Vec<Payment>, Error>;

    /// Returns a transaction object by id
    async fn get_transaction(
        &self,
        transaction_id: &TransactionId,
    ) -> Result<from_db::Transaction, Error>;

    /// Returns a list of a transactions for a given account (and optionally
    /// filter by types). The list of transactions is expected to be sorted by
    /// date desc (newest transactions first)
    async fn get_transactions(
        &self,
        account: &AccountId,
        types: &[Type],
        tags: &[String],
    ) -> Result<Vec<from_db::Transaction>, Error>;

    /// Returns all changelogs associated with a given object_id. The result should be sorted from oldest to newest.
    async fn get_changelogs<T: DeserializeOwned + Serialize + Send + Sync>(
        &self,
        object_id: Vec<u8>,
    ) -> Result<Vec<Changelog<T>>, Error>;
}

#[cfg(test)]
pub mod test {
    use super::*;
    use crate::{payment::SpentInfo, AssetDefinition, AssetManager};
    use rand::Rng;
    use strum::IntoEnumIterator;

    #[macro_export]
    macro_rules! storage_unit_test {
        ($name:ident) => {
            #[tokio::test]
            async fn $name() {
                let assets = AssetManager::new(vec![
                    AssetDefinition::new(1, "BTC", 8),
                    AssetDefinition::new(2, "USD", 4),
                ]);
                let ledger = get_instance(assets.clone(), stringify!($name)).await;
                $crate::storage::test::$name(&ledger, assets).await;
            }
        };
    }

    #[macro_export]
    macro_rules! storage_test_suite {
        () => {
            $crate::storage_unit_test!(transaction);
            $crate::storage_unit_test!(transaction_not_available_until_commit);
            $crate::storage_unit_test!(sorted_unspent_payments);
            $crate::storage_unit_test!(does_not_update_spent_payments);
            $crate::storage_unit_test!(does_not_spend_unspendable_payments);
            $crate::storage_unit_test!(relate_account_to_transaction);
            $crate::storage_unit_test!(pending_new_payments_are_not_spendable);
            $crate::storage_unit_test!(processing_new_payments_are_not_spendable);
            $crate::storage_unit_test!(cancelled_new_payments_are_not_spendable);
            $crate::storage_unit_test!(failed_new_payments_are_not_spendable);
        };
    }

    pub async fn transaction<'a, T>(storage: &'a T, assets: AssetManager)
    where
        T: Storage<'a>,
    {
        let deposit = Transaction::new_external_deposit(
            "test reference".to_owned(),
            Status::Settled,
            vec![(
                "alice".parse().expect("account"),
                assets
                    .human_amount_by_name("USD", "100.99")
                    .expect("valid amount"),
            )],
        )
        .expect("valid tx");
        let mut storing = storage.begin().await.expect("valid tx");
        storing.store_transaction(&deposit).await;
        storing.rollback().await;
        assert!(storage.get_transaction(&deposit.id()).await.is_err());

        let mut storing = storage.begin().await.expect("valid tx");
        storing.store_transaction(&deposit).await;
        storing.commit().await;
        assert!(storage.get_transaction(&deposit.id()).await.is_ok());
    }

    pub async fn transaction_not_available_until_commit<'a, T>(storage: &'a T, assets: AssetManager)
    where
        T: Storage<'a>,
    {
        let deposit = Transaction::new_external_deposit(
            "test reference".to_owned(),
            Status::Settled,
            vec![(
                "alice".parse().expect("account"),
                assets
                    .human_amount_by_name("USD", "100.99")
                    .expect("valid amount"),
            )],
        )
        .expect("valid tx");
        let mut storing = storage.begin().await.expect("valid tx");
        storing.store_transaction(&deposit).await;
        assert!(storage.get_transaction(&deposit.id()).await.is_err());
        storing.rollback().await;
        assert!(storage.get_transaction(&deposit.id()).await.is_err());

        let mut storing = storage.begin().await.expect("valid tx");
        storing.store_transaction(&deposit).await;
        storing.commit().await;
        assert!(storage.get_transaction(&deposit.id()).await.is_ok());
    }

    pub async fn does_not_update_spent_payments<'a, T>(storage: &'a T, assets: AssetManager)
    where
        T: Storage<'a>,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let amount = assets
            .human_amount_by_name("USD", &format!("{}", rng.gen_range(-1000.0..1000.0)))
            .expect("valid amount");

        let payment_id = PaymentId {
            transaction: vec![0u8; 32].try_into().expect("valid tx id"),
            position: 0,
        };

        writer
            .store_new_payment(&Payment {
                id: payment_id.clone(),
                to: account.clone(),
                amount,
                status: Status::Settled,
                changelog: vec![],
                spent: None,
            })
            .await
            .expect("valid payment");

        // Alter state
        assert!(writer
            .update_payment(&payment_id, &payment_id.transaction, Status::Processing)
            .await
            .is_ok());

        // Alter state, again
        assert!(writer
            .update_payment(&payment_id, &payment_id.transaction, Status::Cancelled)
            .await
            .is_ok());

        // Payment is forever used
        assert!(writer
            .update_payment(&payment_id, &payment_id.transaction, Status::Settled)
            .await
            .is_ok());

        for status in Status::iter() {
            assert!(
                writer
                    .update_payment(&payment_id, &payment_id.transaction, status.clone())
                    .await
                    .is_err(),
                "status: {:?}",
                status
            );
        }
    }

    pub async fn pending_new_payments_are_not_spendable<'a, T>(storage: &'a T, assets: AssetManager)
    where
        T: Storage<'a>,
    {
        not_spendable_new_payments_not_spendable(storage, assets, Status::Pending).await
    }

    pub async fn processing_new_payments_are_not_spendable<'a, T>(
        storage: &'a T,
        assets: AssetManager,
    ) where
        T: Storage<'a>,
    {
        not_spendable_new_payments_not_spendable(storage, assets, Status::Processing).await
    }

    pub async fn cancelled_new_payments_are_not_spendable<'a, T>(
        storage: &'a T,
        assets: AssetManager,
    ) where
        T: Storage<'a>,
    {
        not_spendable_new_payments_not_spendable(storage, assets, Status::Cancelled).await
    }

    pub async fn failed_new_payments_are_not_spendable<'a, T>(storage: &'a T, assets: AssetManager)
    where
        T: Storage<'a>,
    {
        not_spendable_new_payments_not_spendable(storage, assets, Status::Failed).await
    }

    pub async fn does_not_spend_unspendable_payments<'a, T>(storage: &'a T, assets: AssetManager)
    where
        T: Storage<'a>,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let amount = assets
            .human_amount_by_name("USD", &format!("{}", rng.gen_range(-1000.0..1000.0)))
            .expect("valid amount");

        for status in Status::iter() {
            if status == Status::Settled {
                continue;
            }
            let status_id: u32 = status.clone().into();
            let payment_id = PaymentId {
                transaction: vec![status_id as u8; 32].try_into().expect("valid tx id"),
                position: 0,
            };

            writer
                .store_new_payment(&Payment {
                    id: payment_id.clone(),
                    to: account.clone(),
                    amount: amount.clone(),
                    status,
                    changelog: vec![],
                    spent: Some(SpentInfo {
                        by: payment_id.transaction.clone(),
                        status: Status::Settled,
                    }),
                })
                .await
                .expect("valid payment");
            for status in Status::iter() {
                if status != Status::Settled {
                    assert!(writer
                        .update_payment(&payment_id, &payment_id.transaction, Status::Pending)
                        .await
                        .is_err());
                }
            }
        }
    }

    pub async fn sorted_unspent_payments<'a, T>(storage: &'a T, assets: AssetManager)
    where
        T: Storage<'a>,
    {
        let mut writer = storage.begin().await.expect("writer");
        let accounts: Vec<AccountId> = (0..10)
            .map(|i| format!("account{}", i).parse().expect("account"))
            .collect();

        let mut rng = rand::thread_rng();
        let mut position = 0;
        let target_inputs_per_account = 20;

        for account in &accounts {
            for i in 0..target_inputs_per_account {
                let amount = assets
                    .human_amount_by_name("USD", &format!("{}", rng.gen_range(-1000.0..1000.0)))
                    .expect("valid amount");
                writer
                    .store_new_payment(&Payment {
                        id: PaymentId {
                            transaction: vec![0u8; 32].try_into().expect("valid tx id"),
                            position,
                        },
                        to: account.clone(),
                        amount,
                        changelog: vec![],
                        status: Status::Settled,
                        spent: None,
                    })
                    .await
                    .expect("valid payment");
                position += 1;
            }
        }
        writer.commit().await.expect("valid commit");
        let mut at_least_one_negative_amount = false;
        for account in &accounts {
            let balance = &storage.get_balance(&account).await.expect("valid balance")[0];
            if balance.cents() < 0 {
                continue;
            }
            let all_unspent_amounts = storage
                .get_unspent_payments(account, balance.asset().id, balance.cents())
                .await
                .expect("valid unspent payments")
                .into_iter()
                .map(|x| x.amount.cents())
                .collect::<Vec<_>>();

            let mut sorted = all_unspent_amounts.clone();
            sorted.sort();

            assert_eq!(target_inputs_per_account, all_unspent_amounts.len());
            assert_eq!(sorted, all_unspent_amounts);
            if !at_least_one_negative_amount {
                at_least_one_negative_amount = sorted[0] < 0;
            }
        }
        assert!(at_least_one_negative_amount);
    }

    pub async fn relate_account_to_transaction<'a, T>(storage: &'a T, assets: AssetManager)
    where
        T: Storage<'a>,
    {
        let account1: AccountId = "alice1".parse().expect("account");
        let account2: AccountId = "alice2".parse().expect("account");
        let account3: AccountId = "alice3".parse().expect("account");
        let deposit = Transaction::new_external_deposit(
            "test reference".to_owned(),
            Status::Settled,
            vec![(
                account1.clone(),
                assets
                    .human_amount_by_name("USD", "100.99")
                    .expect("valid amount"),
            )],
        )
        .expect("valid tx");

        let mut batch = storage.begin().await.expect("valid tx");
        batch.store_transaction(&deposit).await;
        batch
            .relate_account_to_transaction(&deposit, &account1)
            .await
            .expect("relationship");
        batch
            .relate_account_to_transaction(&deposit, &account2)
            .await
            .expect("relationship");

        batch.commit().await.expect("valid commit");

        assert_eq!(
            1,
            storage
                .get_transactions(&account1, &[Type::Deposit], &[])
                .await
                .expect("valid tx")
                .len()
        );
        assert_eq!(
            1,
            storage
                .get_transactions(&account2, &[Type::Deposit], &[])
                .await
                .expect("valid tx")
                .len()
        );
        assert_eq!(
            0,
            storage
                .get_transactions(&account3, &[Type::Deposit], &[])
                .await
                .expect("valid tx")
                .len()
        );
        for account in &[account1, account2, account3] {
            assert_eq!(
                0,
                storage
                    .get_transactions(&account, &[Type::Withdrawal], &[])
                    .await
                    .expect("valid tx")
                    .len()
            );
        }
    }

    async fn not_spendable_new_payments_not_spendable<'a, T>(
        storage: &'a T,
        assets: AssetManager,
        status: Status,
    ) where
        T: Storage<'a>,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let amount = assets
            .human_amount_by_name("USD", &format!("{}", rng.gen_range(10.0..10000.0)))
            .expect("valid amount");

        let payment_id = PaymentId {
            transaction: vec![0u8; 32].try_into().expect("valid tx id"),
            position: 0,
        };

        writer
            .store_new_payment(&Payment {
                id: payment_id.clone(),
                to: account.clone(),
                amount,
                status,
                changelog: vec![],
                spent: None,
            })
            .await
            .expect("valid payment");

        writer.commit().await.is_ok();

        assert!(storage
            .get_balance(&account)
            .await
            .expect("valid balance")
            .is_empty());
    }
}
