use crate::{
    amount::AmountCents,
    changelog::{sort_changes, Changelog},
    payment::{self, SpentInfo},
    storage::{Batch, Storage},
    transaction::*,
    AccountId, Amount, Asset, Payment, PaymentId, Status, TransactionId,
};
use chrono::{serde::ts_milliseconds, DateTime, Utc};
use serde::Serialize;
use sha2::{Digest, Sha256};
use std::{collections::HashMap, ops::Deref};

/// Transactions
///
/// Transactions are the core components of the ledger. The transactions are a
/// list of unspent payments that are about to be spend, to create a new set of
/// Payments, that can be spend in the future. This model is heavily inspired in
/// Bitcoin's UTXO model. The terms in this context are payments, spend and
/// create instead of unspent transactions, input and output.
///
/// This simple architecture allows to track accounts pretty efficiently,
/// because all that matters are unspent payments owned by a given account.
/// Every spent payment is stored for historical reasons but it is not relevant
/// for any calculations regarding available funds.
///
/// The transaction has a few rules, for instance the sum of spend Payments
/// should be the same as create Payments, for each easy. There is no 'fee'
/// concept, so any mismatch in any direction will error the constructor.
///
/// Transactions are immutable after they are finalized, and the payments can
/// only be re-usable if the transaction failed or was cancelled. Once the
/// transaction settles the spent payments are forever spent. Any rollback
/// should be a new transaction, initiated by a higher layer.
///
/// The spent payments are unavailable until the transaction is finalized,
/// either as settled, cancelled or failed. A higher layer should split any
/// available payment to be spend into a new transaction, and then finalize the
/// transaction, and reserve only the exact amount to be spent, otherwise
/// unrelated funds will be held unspendable until the transaction is finalized.
#[derive(Debug, Clone, Serialize)]
pub struct Transaction {
    id: TransactionId,
    #[serde(skip_serializing_if = "Vec::is_empty")]
    spends: Vec<Payment>,
    #[allow(dead_code)]
    reference: String,
    #[serde(rename = "type")]
    typ: Type,
    #[serde(skip_serializing_if = "Vec::is_empty")]
    creates: Vec<Payment>,
    changelog: Vec<Changelog<ChangelogEntry>>,
    status: Status,
    #[serde(skip_serializing_if = "Vec::is_empty")]
    tags: Vec<String>,
    #[serde(with = "ts_milliseconds")]
    created_at: DateTime<Utc>,
    #[serde(with = "ts_milliseconds")]
    updated_at: DateTime<Utc>,
}

impl Transaction {
    /// Creates a new external deposit transaction
    ///
    /// All transactions must be balanced, same amounts that are spent should be
    /// created. There are two exceptions, external deposits and withdrawals.
    /// The idea is to mimic external operations, where new assets enter the system.
    pub fn new_external_deposit(
        reference: String,
        status: Status,
        pay_to: Vec<(AccountId, Amount)>,
    ) -> Result<Transaction, Error> {
        let created_at = Utc::now();
        let id = Self::calculate_id(
            &reference,
            vec![],
            pay_to
                .iter()
                .map(|t| (&t.0, &t.1))
                .collect::<Vec<(&AccountId, &Amount)>>(),
            created_at,
        )?;
        let create = pay_to
            .into_iter()
            .enumerate()
            .map(|(position, (to, amount))| {
                let id = crate::PaymentId {
                    transaction: id.clone(),
                    position,
                };
                let id_bytes = id.bytes().to_vec();
                Payment {
                    id,
                    to,
                    amount,
                    spent: None,
                    changelog: vec![Changelog::new(
                        None,
                        id_bytes,
                        payment::ChangelogEntry {
                            status: status.clone(),
                            reason: reference.clone(),
                            spent: None,
                        },
                    )],
                    status: status.clone(),
                }
            })
            .collect();

        let changelog: Vec<Changelog<changelog::ChangelogEntry>> = vec![Changelog::new(
            None,
            id.deref().to_vec(),
            ChangelogEntry {
                status: status.clone(),
                reason: reference.clone(),
            },
        )];

        Ok(Self {
            id,
            spends: vec![],
            creates: create,
            reference,
            typ: Type::Deposit,
            tags: Vec::new(),
            status,
            changelog,
            created_at,
            updated_at: Utc::now(),
        })
    }

    /// Returns a mutable reference to the tags associated with this transaction
    pub fn get_tags_mut(&mut self) -> &mut [String] {
        &mut self.tags
    }

    /// Creates a new external withdrawal transaction
    ///
    /// Burns assets to reflect external withdrawals
    pub fn new_external_withdrawal(
        reference: String,
        status: Status,
        spend: Vec<Payment>,
    ) -> Result<Transaction, Error> {
        let created_at = Utc::now();
        let id = Self::calculate_id(
            &reference,
            spend.iter().map(|t| &t.id).collect::<Vec<&PaymentId>>(),
            vec![],
            created_at,
        )?;
        let spend = spend
            .into_iter()
            .map(|mut payment| {
                payment.spent = Some(SpentInfo {
                    by: id.clone(),
                    status: status.clone(),
                });
                payment.new_changelog(&reference);
                payment
            })
            .collect();

        let changelog: Vec<Changelog<changelog::ChangelogEntry>> = vec![Changelog::new(
            None,
            id.deref().to_vec(),
            ChangelogEntry {
                status: status.clone(),
                reason: reference.clone(),
            },
        )];

        Ok(Self {
            id,
            spends: spend,
            creates: vec![],
            typ: Type::Withdrawal,
            tags: Vec::new(),
            reference,
            status,
            changelog,
            created_at,
            updated_at: Utc::now(),
        })
    }

    /// Creates a new transaction
    pub async fn new(
        reference: String,
        status: Status,
        typ: Type,
        spend: Vec<Payment>,
        pay_to: Vec<(AccountId, Amount)>,
    ) -> Result<Transaction, Error> {
        let created_at = Utc::now();
        let id = Self::calculate_id(
            &reference,
            spend.iter().map(|t| &t.id).collect::<Vec<&PaymentId>>(),
            pay_to
                .iter()
                .map(|t| (&t.0, &t.1))
                .collect::<Vec<(&AccountId, &Amount)>>(),
            created_at,
        )?;

        for (i, input) in spend.iter().enumerate() {
            if !input.is_spendable_or_was_by(&id) {
                return Err(Error::InvalidPaymentStatus(i, input.status.clone()));
            }
        }
        let spend = spend
            .into_iter()
            .map(|mut payment| {
                payment.spent = Some(SpentInfo {
                    by: id.clone(),
                    status: status.clone(),
                });
                payment.new_changelog(&reference);
                payment
            })
            .collect();

        let create = pay_to
            .into_iter()
            .enumerate()
            .map(|(position, (to, amount))| {
                let mut payment = Payment {
                    id: crate::PaymentId {
                        transaction: id.clone(),
                        position,
                    },
                    to,
                    amount,
                    spent: None,
                    changelog: vec![],
                    status: status.clone(),
                };
                payment.new_changelog(&reference);
                payment
            })
            .collect();

        let changelog: Vec<Changelog<changelog::ChangelogEntry>> = vec![Changelog::new(
            None,
            id.deref().to_vec(),
            ChangelogEntry {
                status: status.clone(),
                reason: reference.clone(),
            },
        )];

        Ok(Self {
            id,
            reference,
            spends: spend,
            typ,
            creates: create,
            tags: Vec::new(),
            status,
            changelog,
            created_at,
            updated_at: Utc::now(),
        })
    }

    fn calculate_id(
        reference: &str,
        spend: Vec<&PaymentId>,
        create: Vec<(&AccountId, &Amount)>,
        created_at: DateTime<Utc>,
    ) -> Result<TransactionId, Error> {
        let mut hasher = Sha256::new();
        let mut spend = spend;

        spend.sort();

        for id in spend.into_iter() {
            hasher.update(&bincode::serialize(id)?);
        }
        for (account, amount) in create.into_iter() {
            hasher.update(&bincode::serialize(account)?);
            hasher.update(&bincode::serialize(amount)?);
        }
        hasher.update(created_at.timestamp_millis().to_le_bytes());
        hasher.update(reference);
        Ok(TransactionId::new(hasher.finalize().into()))
    }

    /// Settles the current transaction
    ///
    /// This is equivalent to changes the status to Settle and to persist
    pub async fn settle<'a, S>(&mut self, storage: &'a S, reason: String) -> Result<(), Error>
    where
        S: Storage<'a> + Sync + Send,
    {
        self.change_status(Status::Settled, reason)?;
        self.persist::<S>(storage).await
    }

    /// Changes the status of a given transaction
    #[inline]
    pub fn change_status(&mut self, new_status: Status, reason: String) -> Result<(), Error> {
        if self.status.can_transition_to(&new_status) {
            for payment in self.spends.iter_mut() {
                payment.spent = if new_status.is_rollback() {
                    None
                } else {
                    Some(SpentInfo {
                        by: self.id.clone(),
                        status: new_status.clone(),
                    })
                };
                payment.new_changelog(&reason);
            }

            for payment in self.creates.iter_mut() {
                payment.status = new_status.clone();
                payment.new_changelog(&reason);
            }

            let previous = self.changelog.last().map(|x| x.id()).transpose()?;
            self.changelog.push(Changelog::new(
                previous,
                self.id.deref().to_vec(),
                ChangelogEntry {
                    status: new_status.clone(),
                    reason,
                },
            ));
            self.status = new_status;
            Ok(())
        } else {
            Err(Error::StatusTransitionNotAllowed(
                self.status.clone(),
                new_status,
            ))
        }
    }

    #[inline]
    /// Checks if the transaction attempts to spend any negative amount (which is no allowed)
    fn check_no_negative_amounts_are_spent(
        &self,
        debit: &HashMap<Asset, i128>,
    ) -> Result<(), Error> {
        for (asset, amount) in debit.iter() {
            if *amount <= 0 {
                return Err(Error::InvalidAmount(
                    asset.new_amount(*amount),
                    asset.new_amount(*amount),
                ));
            }
        }

        Ok(())
    }

    /// Validates the transaction before storing
    pub(crate) fn validate(&self) -> Result<(), Error> {
        let calculated_id = Self::calculate_id(
            &self.reference,
            self.spends.iter().map(|p| &p.id).collect::<Vec<_>>(),
            self.creates
                .iter()
                .map(|p| (&p.to, &p.amount))
                .collect::<Vec<_>>(),
            self.created_at,
        )?;

        if calculated_id != self.id {
            return Err(Error::InvalidTransactionId(self.id.clone(), calculated_id));
        }

        let mut debit = HashMap::<Asset, AmountCents>::new();
        let mut credit = HashMap::<Asset, AmountCents>::new();

        for (i, input) in self.spends.iter().enumerate() {
            if !input.is_spendable_or_was_by(&self.id) {
                return Err(Error::SpentPayment(i));
            }
            if let Some(value) = debit.get_mut(input.amount.asset()) {
                *value = input
                    .amount
                    .cents()
                    .checked_add(*value)
                    .ok_or(Error::Overflow)?;
            } else {
                debit.insert(input.amount.asset().clone(), input.amount.cents());
            }
        }

        self.check_no_negative_amounts_are_spent(&debit)?;

        if !self.typ.is_transaction() {
            // We don't care input/output balance in external operations
            // (withdrawals/deposits), because these operations are inbalanced
            return Ok(());
        }

        for output in self.creates.iter() {
            if let Some(value) = credit.get_mut(output.amount.asset()) {
                *value = output
                    .amount
                    .cents()
                    .checked_add(*value)
                    .ok_or(Error::Overflow)?;
            } else {
                credit.insert(output.amount.asset().clone(), output.amount.cents());
            }
        }

        for (asset, credit_amount) in credit.into_iter() {
            if let Some(debit_amount) = debit.remove(&asset) {
                if debit_amount != credit_amount {
                    return Err(Error::InvalidAmount(
                        asset.new_amount(debit_amount),
                        asset.new_amount(credit_amount),
                    ));
                }
            } else {
                return Err(Error::MissingSpendingAsset(asset));
            }
        }

        if let Some((asset, _)) = debit.into_iter().next() {
            return Err(Error::MissingPaymentAsset(asset));
        }

        Ok(())
    }

    /// Returns the list of payments that were used to create this transaction
    pub fn spends(&self) -> &[Payment] {
        &self.spends
    }

    /// Returns the list of payments that were created by this transaction
    pub fn creates(&self) -> &[Payment] {
        &self.creates
    }

    /// Returns the transaction ID
    pub fn id(&self) -> &TransactionId {
        &self.id
    }

    /// Returns the transaction status
    pub fn status(&self) -> &Status {
        &self.status
    }

    /// Returns the transaction type
    pub fn typ(&self) -> &Type {
        &self.typ
    }

    /// Returns the reference of this transaction
    pub fn reference(&self) -> &str {
        &self.reference
    }

    /// Returns the last version of this transaction (the ID of the last entry from the changelog)
    pub fn last_version(&self) -> Option<Vec<u8>> {
        if let Some(Ok(x)) = self.changelog.last().map(|x| x.id()) {
            Some(x)
        } else {
            None
        }
    }

    /// Returns the time when this transaction was created
    pub fn created_at(&self) -> DateTime<Utc> {
        self.created_at
    }

    /// Returns the time when this transaction was last updated
    pub fn updated_at(&self) -> DateTime<Utc> {
        self.updated_at
    }

    /// Persists the changes done to this transaction object.
    /// This method is not idempotent, and it will fail if the transaction if the requested update is not allowed.
    pub async fn persist<'a, S>(&mut self, storage: &'a S) -> Result<(), Error>
    where
        S: Storage<'a> + Sync + Send,
    {
        let mut batch = storage.begin().await?;
        if let Some(status) = batch.get_payment_status(&self.id).await? {
            if status.is_finalized() {
                return Err(Error::TransactionUpdatesNotAllowed);
            }
        }
        self.validate()?;
        self.updated_at = Utc::now();
        batch.store_transaction(self).await?;
        for payment in self.creates.iter() {
            batch.store_new_payment(payment).await?;
            batch
                .relate_account_to_transaction(self, &payment.to)
                .await?;
            batch.store_changelogs(&payment.changelog).await?;
        }
        for input in self.spends.iter() {
            batch
                .update_payment(&input.id, &self.id, self.status.clone())
                .await?;
            batch.relate_account_to_transaction(self, &input.to).await?;
            batch.store_changelogs(&input.changelog).await?;
        }

        batch.tag_transaction(self, &self.tags).await?;
        batch.store_changelogs(&self.changelog).await?;
        batch.commit().await?;
        Ok(())
    }
}

impl TryFrom<from_db::Transaction> for Transaction {
    type Error = Error;

    fn try_from(value: from_db::Transaction) -> Result<Self, Self::Error> {
        let tx = Transaction {
            id: value.id,
            typ: value.typ,
            spends: value.spend,
            creates: value.create,
            reference: value.reference,
            tags: value.tags,
            status: value.status,
            changelog: sort_changes(value.changelog, value.last_change)?,
            created_at: value.created_at,
            updated_at: value.updated_at,
        };
        tx.validate()?;
        Ok(tx)
    }
}
